


/*********************************************************************
 * (C) Copyright 2010 INRIA Nancy Grand Est
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *********************************************************************/









/*********************************************************************
 *
 * File: search.c
 *
 * Description: implementation of routines that search the "state" space
 *
 *
 * Author: Joerg Hoffmann 2004
 *
 *********************************************************************/ 









#include "ff.h"

#include "output.h"
#include "memory.h"

#include "relax.h"
#include "search.h"
















/************************************
 * AND-OR SEARCH IMPLEMENTATION *
 ************************************/

























/*
 * the use of h fn weight in here is right now a tad simplistic:
 *       we weight the leaves only, ie the direct h fn values
 *       those are multiplied with w
 * 
 * according to LAO* paper, the more precise approach is to instead
 * propagate h and g parts of f value separately, where leaf g is 0 
 * and leaf h is h(s); then the updates go g := 1 + sum/max_sons g(son) 
 * respectively sum/max_sons h(son). then f := g + w*h as usual.
 *
 * ... maybe implement at some point.. although I doubt it's gonna
 *  make much of a difference.
 */


void do_and_or_search( void )

{

  static Bool fc = TRUE;
  static State S;
  static int *A;

  BfsNode *open_leaf, *tmpbfs;
  int  i, j, k, minlf = INFINITY, minif = INFINITY, maxd = -1;
  int  num_A, prevf, propf;
  Bool start = TRUE;
  BfsEdge *tmpedge;



  if ( fc ) {
    make_state( &S, gnum_ft_conn );
    A = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
    fc = FALSE;
  }



  /* create initial search node; first, its (single-leaf-node) plantree
   */
  gbfs_initial_state = new_BfsNode();
  make_state( &(gbfs_initial_state->S), ginitial_state.num_F );
  source_to_dest( &(gbfs_initial_state->S), &(ginitial_state) );
  if ( contains_goal( &(gbfs_initial_state->S) ) ) {
    if ( gcmd_line.debug >= 1 ) {
      printf("\ninitial state contains goal!");
    }
    gbfs_initial_state->solved = TRUE;
    gbfs_initial_state->f = 0;
  } else {
    if ( gcmd_line.helpful ) {
      gbfs_initial_state->f = get_1P_and_H( gbfs_initial_state );
      gbfs_initial_state->H = ( int * ) calloc( gnum_H, sizeof( int ) );
      for ( i = 0; i < gnum_H; i++ ) {
	gbfs_initial_state->H[i] = gH[i];
      }
      gbfs_initial_state->num_H = gnum_H;
    } else {
      gbfs_initial_state->f = get_1P( gbfs_initial_state );
    }
    if ( gcmd_line.debug >= 1 ) {
      printf("\ninitial state 1P: %d, weighted %d", 
	     gbfs_initial_state->f, gcmd_line.hweight * gbfs_initial_state->f);
    }
  }
  if ( gbfs_initial_state->f == INFINITY ) {
    if ( gcmd_line.debug >= 1 ) {
      printf("\ninitial h value infty!");
    }
    gbfs_initial_state->failed = TRUE;
  } else {
    gbfs_initial_state->f *= gcmd_line.hweight;
  }
  gbfs_initial_state->depth = 0;



  /* search: expand nongoal leaf of current best subtree, until
   * the best subtree has no nongoal leaf -> is a plan
   */
  while ( TRUE ) {

    /* plan existence and non-existence are propagated through
     * the search space when they are detected.
     */
    if ( gbfs_initial_state->solved ||
	 gbfs_initial_state->failed ) {
      break;
    }

    /* none of both -> find an open leaf of currently best subtree
     */
    if ( !gcmd_line.allow_failed_nodes ) {
      open_leaf = get_open_leaf_of_best_subtree();
    } else {
      open_leaf = SAM_get_open_leaf_of_best_subtree();
    }
    if ( open_leaf->depth == MAX_PLAN_LENGTH ) {
      printf("\nincrease MAX_PLAN_LENGTH! currently %d\n\n", MAX_PLAN_LENGTH);
      exit( 1 );
    }

    /* screen printout
     */
    if ( LESS( open_leaf->f, minlf ) ||
	 LESS( gbfs_initial_state->f, minif ) ||
	 maxd < open_leaf->depth ) {
      if ( LESS( open_leaf->f, minlf ) ) {
	minlf = open_leaf->f;
      }
      if ( LESS( gbfs_initial_state->f, minif ) ) {
	minif = gbfs_initial_state->f;
      }
      if ( maxd < open_leaf->depth ) {
	maxd = open_leaf->depth;
      }
      if ( start ) {
	printf("\nmin leaf goaldist/ini state f: %3d/%3d; maxdepth: %3d", 
	       minlf, minif, maxd);
	start = FALSE;
	fflush(stdout);
      } else {
	printf("\n                               %3d/%3d            %3d", 
	       minlf, minif, maxd);
	fflush(stdout);
      }
      gmax_search_depth = maxd;
    }

    /* expand the leaf
     */
    if ( gcmd_line.debug >= 1 ) {
      printf("\nopen leaf:");
      print_state( open_leaf->S );
    }
    if ( gcmd_line.debug >= 2 ) {
      printf("\nused nondet actions: %d\n", 
	     open_leaf->num_used_nondets);
      print_used_nondets( open_leaf );
    }

    /* with either all applicable actions, or the helpful actions
     */
    if ( gcmd_line.helpful ) {
      for ( i = 0; i < open_leaf->num_H; i++ ) {
	A[i] = open_leaf->H[i];
      }
      num_A = open_leaf->num_H;
    } else {
      get_A( open_leaf );
      for ( i = 0; i < gnum_A; i++ ) {
	A[i] = gA[i];
      }
      num_A = gnum_A;
    }

    /* now apply all the A actions
     */
    for ( i = 0; i < num_A; i++ ) {
      if ( gcmd_line.debug >= 1 ) {
	printf("\napplying action ");
	print_op_name( A[i] );
	printf(", number of outcomes: %d", 
	       gop_conn[A[i]].num_E);
      }

      /* apply each possible outcome of the action
       *
       * we will have 1 "edge" ie openleaf --> edge 
       * and edge --> n for all new bfsnodes n, i.e.,
       * for all outcomes.
       *
       * the new edge has number out edges initialized to 0,
       * in memory.c
       */

      /* before we do anything: 
       *
       * check if we had this guy already, on a path
       * of predecessors up to the closest non-trivial
       * AND node
       *
       * makes sense of course only if this guy itself
       * is trivial.
       */
      if ( gcmd_line.simple_duplicates ) {
	if ( gop_conn[A[i]].num_E == 1 ) {
	  result_to_dest( &S, &(open_leaf->S), gop_conn[A[i]].E[0]);
	  if ( gcmd_line.debug >= 2 ) {
	    printf("\nchecking for duplicate of: ");
	    print_state( S );
	  }
	  if ( has_duplicate_simple_predecessor( &S, open_leaf ) ) {
	    /* skip this action!
	     */
	    continue;
	  }
	}
      }

      /* Ok, duplicate check does not apply. Let's go.
       */
      tmpedge = new_BfsEdge();
      tmpedge->op = A[i];
      tmpedge->in_node = open_leaf;
      tmpedge->next = open_leaf->outgoing_edges;
      open_leaf->outgoing_edges = tmpedge;
      for ( j = 0; j < gop_conn[A[i]].num_E; j++ ) {
	if ( gcmd_line.debug >= 2 ) {
	  printf("\napplying action ");
	  print_op_name( A[i] );
	  printf(", outcome number %d", j);
	}
	tmpbfs = new_BfsNode();

	/* establish the updated set of used non-det actions.
	 *
	 * if the current action is det, the set remains the same.
	 * else, the current action is added into it.
	 */
	if ( gop_conn[A[i]].num_E <= 1 ) {
	  if ( open_leaf->num_used_nondets > 0 ) {
	    tmpbfs->used_nondets = ( int * ) 
	      calloc( open_leaf->num_used_nondets, sizeof( int ) );
	  }
	} else {
	  tmpbfs->used_nondets = ( int * ) 
	    calloc( open_leaf->num_used_nondets + 1 + gop_conn[A[i]].num_DNF, 
		    sizeof( int ) );
	}
	for ( k = 0; k < open_leaf->num_used_nondets; k++ ) {
	  tmpbfs->used_nondets[k] = open_leaf->used_nondets[k];
	}
	tmpbfs->num_used_nondets = open_leaf->num_used_nondets;
	if ( gop_conn[A[i]].num_E > 1 ) {
	  tmpbfs->used_nondets[tmpbfs->num_used_nondets] = A[i];
	  tmpbfs->num_used_nondets++;
	  /* need to exclude also all actions that were generated from
	   * the same operator! ---> the DNF brothers.
	   */
	  for ( k = 0; k < gop_conn[A[i]].num_DNF; k++ ) {
	    tmpbfs->used_nondets[tmpbfs->num_used_nondets] = gop_conn[A[i]].DNF[k];
	    tmpbfs->num_used_nondets++;
	  }
	}
	if ( gcmd_line.debug >= 2 ) {
	  printf("\nnew used nondet actions: %d\n", 
		 tmpbfs->num_used_nondets);
	  print_used_nondets( tmpbfs );
	}


	/* now insert this into the search space graph structure
	 */
	tmpbfs->depth = open_leaf->depth + 1;
	tmpbfs->ingoing_edge = tmpedge;
	tmpedge->out_nodes[tmpedge->num_out_nodes] =
	  tmpbfs;
	tmpedge->num_out_nodes++;
 
	/* compute the state...
	 */
	result_to_dest( &S, &(open_leaf->S), gop_conn[A[i]].E[j]);
	if ( gcmd_line.debug >= 2 ) {
	  printf("\nresult state:");
	  print_state( S );
	}

	/* ... and insert that state as well as its h values
	 * into the node.
	 */
	set_state_and_hvalues( &S, tmpbfs );

      } /* end of loop over all outcomes */
    } /* for all A applicable actions */
    


    /* re-adjust the f values and the pointers to best edges, 
     * as well as solved and failed values,
     * along path from expanded leaf back to initial state
     *
     * TWO versions: one corresponding to the usual AO*
     *               one corresponding to the SAM semantics as per
     *               SPARK'09, where failed nodes are allowed
     *               below ANDs. 
     */
    if ( !gcmd_line.allow_failed_nodes ) {
      update_fvalues_and_markers( open_leaf );
    } else {
      SAM_update_fvalues_and_markers( open_leaf );
    }
  }



  if ( gbfs_initial_state->solved ) {
    if ( !verify_plan() ) {
      ginvalid_plan = TRUE;
      printf("\n\nWARNING: plan verification failed! unsolvable leaf with finite h value! Helpful actions pruning might have caused plan to be cut short!\n\n");
    }
    print_plan();
  } else {
    if ( !gcmd_line.helpful ) {
      printf("\n\ninitial state failed! problem proved unsolvable!\n\n");
    } else {
      printf("\n\ninitial state failed! try without helpful actions pruning!\n\n");
    }
  }

}



BfsNode *get_open_leaf_of_best_subtree( void )

{

  BfsNode *i_bfs;
  int i;

  /* simply follow non-solved best path, taking
   * first out-edge that appears. (ie no selection heuristic here)
   *
   * NOTE: assumes there are no cycles! (think about that
   * in detail when doing more than stagnating repeated-checks)
   */
  
  i_bfs = gbfs_initial_state;
  while ( i_bfs ) {
    if ( i_bfs->best_outgoing_edge == NULL ) break;
    for ( i = 0; 
	  i < i_bfs->best_outgoing_edge->num_out_nodes; 
	  i++ ) {
      if ( !i_bfs->best_outgoing_edge->out_nodes[i]->solved ) {
	i_bfs = i_bfs->best_outgoing_edge->out_nodes[i];
	if ( i_bfs->failed ) {
	  printf("\n\n2 failed in open leaf getting?\n\n");
	  exit( 1 );
	}
	break;
      }
    } /* endfor over the outgoing edges */
  }

  return i_bfs;

}



BfsNode *SAM_get_open_leaf_of_best_subtree( void )

{

  BfsNode *i_bfs;
  int i;

  /* simply follow non-solved best path, taking
   * first out-edge that appears. (ie no selection heuristic here)
   *
   * NOTE: assumes there are no cycles! (think about that
   * in detail when doing more than stagnating repeated-checks)
   */
  
  i_bfs = gbfs_initial_state;
  while ( i_bfs ) {
    if ( i_bfs->best_outgoing_edge == NULL ) break;
    for ( i = 0; 
	  i < i_bfs->best_outgoing_edge->num_out_nodes; 
	  i++ ) {
      /* in SAM, it may happen that some states attached to the
       * best action are failed!!
       */
      if ( !i_bfs->best_outgoing_edge->out_nodes[i]->solved &&
	   !i_bfs->best_outgoing_edge->out_nodes[i]->failed ) {
	i_bfs = i_bfs->best_outgoing_edge->out_nodes[i];
	if ( i_bfs->failed ) {
	  printf("\n\n2 failed in open leaf getting?\n\n");
	  exit( 1 );
	}
	break;
      }
    } /* endfor over the outgoing edges */
  }

  return i_bfs;

}



void set_state_and_hvalues( State *S, BfsNode *tmpbfs )

{

  int i;

  /* make space for, and set, the state.
   */
  make_state( &(tmpbfs->S), S->num_F );
  source_to_dest( &(tmpbfs->S), S );

  /* check if this is a goal state.
   */
  if ( contains_goal( &(tmpbfs->S) ) ) {
    /* mark this node as solved!
     */
    if ( gcmd_line.debug >= 1 ) {
      printf("\ncontains goal!");
    }
    tmpbfs->solved = TRUE;
    tmpbfs->f = 0;

    return;
  }

  /* get h value and helpful actions.
   */
  if ( gcmd_line.helpful ) {
    tmpbfs->f = get_1P_and_H( tmpbfs );
    tmpbfs->H = ( int * ) calloc( gnum_H, sizeof( int ) );
    for ( i = 0; i < gnum_H; i++ ) {
      tmpbfs->H[i] = gH[i];
    }
    tmpbfs->num_H = gnum_H;
  } else {
    tmpbfs->f = get_1P( tmpbfs );
  }
  if ( gcmd_line.debug >= 1 ) {
    printf("\n1P: %d, weighted: %d", 
	   tmpbfs->f, tmpbfs->f * gcmd_line.hweight);
  }

  /* accordingly set the values in the node.
   */
  if ( tmpbfs->f == INFINITY ) {
    /* mark this node as failed!
     */
    if ( gcmd_line.debug >= 1 ) {
      printf("\nh value infty!");
    }    tmpbfs->failed = TRUE;
  } else {
    tmpbfs->f *= gcmd_line.hweight;
  }

}



void update_fvalues_and_markers( BfsNode *open_leaf )

{

  BfsNode *i_bfs;
  BfsEdge *i_edge, *bestedge;
  int  minf, f;
  Bool ex_solved, ex_open;
  int num_solved_sons, num_open_sons, num_failed_sons;
  int i;



  i_bfs = open_leaf;
  while ( TRUE ) {

    /* adjust the f value of the node, and the best edge,
     * and the solved and failed values.
     */

    /* these are the values we wish to compute, for this node.
     * (min is over possible actions)
     */
    minf = -2;
    bestedge = NULL;
    ex_open = FALSE;
    ex_solved = FALSE;
    for ( i_edge = i_bfs->outgoing_edges; i_edge; i_edge = i_edge->next ) {

      num_open_sons = 0;
      num_solved_sons = 0;
      num_failed_sons = 0;
      f = -2;
      for ( i = 0; i < i_edge->num_out_nodes; i++ ) {
	if ( i_edge->out_nodes[i]->failed ) {
	  if ( i_edge->out_nodes[i]->f != INFINITY ) {
	    printf("\nfailed node with finite f value?\n\n");
	    exit( 1 );
	  }
	  /* ... this is definitely the maximum f-value...
	   */
	  f = INFINITY;
	  num_failed_sons++;
	  continue;
	}
	if ( i_edge->out_nodes[i]->f == INFINITY ) {
	  printf("\nunfailed node with infinite f value?\n\n");
	  exit( 1 );
	}
	if ( i_edge->out_nodes[i]->f < 0 ) {
	  printf("\n2 unfailed node with < 0 f value?\n\n");
	  exit( 1 );
	}

	if ( i_edge->out_nodes[i]->solved ) {
	  num_solved_sons++;
	} else {
	  num_open_sons++;
	}

	/* SAP-FF: take f-values to be the maximum, in AND nodes.
	 * this corresponds to minimizing depth of the solution tree.
	 *
	 * in practice, ie contigent-ff, this turned out to be WAY
	 * faster than taking the sum (the h fn is more akin to tree
	 * depth than to tree size)
	 *
	 * NOTE: maximum here is, with the above, over ALL
	 * successors, includig the failed ones.
	 * NOTE also that we use the LESS function, which
	 * covers the case where f is already infinite.
	 * (else, since infty is encoded as -1, there could
	 * be misunderstandings here...)
	 */
	if ( f == -2 || LESS( f, i_edge->out_nodes[i]->f ) ) {
	  f = i_edge->out_nodes[i]->f;
	}
      } /* endfor i over the outnodes of this edge */



      /* now minimize the f values over the possible actions, i.e.,
       * the outgoing edges of the bfs node.
       *
       * Note that we use the LESS function!
       * Hence, if f is infinity (which is encoded as -1), 
       *     it will not be taken as the new minimum!
       *
       * NOTE: if f == -2 then this action had no outcome states. 
       * this cannot happen. just debug to be fool-ptoof sure.
       */
      if ( f == -2 ) {
	printf("\nf == -2 in non-SAM update markers?\n\n");
	exit( 1 );
      }
      if ( minf == -2 || LESS( f, minf ) ) {
	minf = f;
	bestedge = i_edge;
      }

      /* debugging... just to be fool-proof sure...
       */
      if ( num_failed_sons + num_open_sons + num_solved_sons 
	   != i_edge->num_out_nodes ) {
	printf("\ntotal number of sons counted wrongly?\n\n");
	exit( 1 );
      }

      /* ... and remember whether one of these guys was solved
       * or open.
       *
       * the latter is true iff no son is failed, and there is at least
       * one open son.
       *
       * the former is true iff all sons are solved, and there is at least
       * one son.
       */
      if ( num_failed_sons == 0 && num_open_sons > 0 ) {
	ex_open = TRUE;
      }
      if ( num_solved_sons == i_edge->num_out_nodes && 
	   num_solved_sons > 0  ) {
	ex_solved = TRUE;
      }

    } /* loop over all outgoing edges i_edge */



    /* make the "OR" over the state node.
     */
    if ( !ex_open && !ex_solved ) {
      i_bfs->failed = TRUE;
      i_bfs->f = INFINITY;
    } else {
      if ( ex_solved ) {
	i_bfs->solved = TRUE;
      }
      /* debug...
       */
      if ( minf == -2 || bestedge == NULL ) {
	printf("\n\nmin/bestedge not set?\n\n");
	exit( 1 );
      }    
      i_bfs->f = 1 + minf;
      i_bfs->best_outgoing_edge = bestedge;
    }



    /* go 1 more step towards the root of the tree
     */
    if ( i_bfs->ingoing_edge != NULL ) {
      i_bfs = i_bfs->ingoing_edge->in_node;
    } else {
      break;
    }

  } /* while loop backwards over bfs nodes from open leaf */

}



void SAM_update_fvalues_and_markers( BfsNode *open_leaf )

{

  BfsNode *i_bfs;
  BfsEdge *i_edge, *bestedge;
  int  minf, f;
  Bool ex_solved, ex_open;
  int num_solved_sons, num_open_sons, num_failed_sons;
  int i;



  i_bfs = open_leaf;
  while ( TRUE ) {

    /* adjust the f value of the node, and the best edge,
     * and the solved and failed values.
     */

    /* these are the values we wish to compute, for this node.
     * (min is over possible actions)
     */
    minf = -2;
    bestedge = NULL;
    ex_open = FALSE;
    ex_solved = FALSE;
    if ( gcmd_line.debug > 2 ) 
      printf("\nSAM: starting state");
    for ( i_edge = i_bfs->outgoing_edges; i_edge; i_edge = i_edge->next ) {
      if ( gcmd_line.debug > 2 ) 
	printf("\nSAM: new action");

      num_open_sons = 0;
      num_solved_sons = 0;
      num_failed_sons = 0;
      f = -2;
      for ( i = 0; i < i_edge->num_out_nodes; i++ ) {
	if ( gcmd_line.debug > 2 ) printf("\nSAM: new action outcome state");
	if ( i_edge->out_nodes[i]->failed ) {
	  if ( i_edge->out_nodes[i]->f != INFINITY ) {
	    printf("\nSAM failed node with finite f value?\n\n");
	    exit( 1 );
	  }
	  if ( gcmd_line.debug > 2 ) 
	    printf("\nSAM: failed son f value %d", i_edge->out_nodes[i]->f);
	  num_failed_sons++;
	  continue;
	}
	if ( i_edge->out_nodes[i]->f == INFINITY ) {
	  printf("\nSAM unfailed node with infinite f value?\n\n");
	  exit( 1 );
	}
	if ( i_edge->out_nodes[i]->f < 0 ) {
	  printf("\nSAM unfailed node with < 0 f value?\n\n");
	  exit( 1 );
	}

	if ( i_edge->out_nodes[i]->solved ) {
	  num_solved_sons++;
	  if ( gcmd_line.debug > 2 ) 
	    printf("\nSAM: solved son f value %d", i_edge->out_nodes[i]->f);
	} else {
	  num_open_sons++;
	  if ( gcmd_line.debug > 2 ) 
	    printf("\nSAM: open son f value %d", i_edge->out_nodes[i]->f);
	}

	/* SAP-FF: take f-values to be the maximum, in AND nodes.
	 * this corresponds to minimizing depth of the solution tree.
	 *
	 * in practice, ie contigent-ff, this turned out to be WAY
	 * faster than taking the sum (the h fn is more akin to tree
	 * depth than to tree size)
	 *
	 * NOTE: here the maximum is taken over the non-failed
	 * successor only! because failed successors are considered
	 * to be "solved". --> above, f = infty is NOT set, here.
	 */
	if ( f == -2 || i_edge->out_nodes[i]->f > f ) {
	  if ( gcmd_line.debug > 2 ) 
	    printf("\nSAM: update f value to %d", i_edge->out_nodes[i]->f);
	  f = i_edge->out_nodes[i]->f;
	}
      } /* endfor i over the outnodes of this edge */


      /* ACHTUNG!  this way of selecting the "best action", ie the min
       * f, may change the previous best action although the new one
       * is only same f-value, not strictly better. 
       *
       * Example: two acts, a1 and a2. first a2 looks better but then
       * turns out need more steps than what h suggested. come back to 
       * this node in updates, and set best act to a1 now even if both
       * f-values are actually still the same. 
       *
       * thus the search may be more "breadth-style" in the sense that
       * equally good-looking options are swithced. perhaps it would
       * be better for efficiency (eg in the presence of
       * transpositions, eg when we have several BOs), to disallow
       * such behavior?
       */


      /* now minimize the f values over the possible actions, i.e.,
       * the outgoing edges of the bfs node.
       *
       * Note that we use the LESS function!
       * Hence, if f is infinity (which is encoded as -1), 
       *     it will not be taken as the new minimum!
       *
       * NOTE: actually, here f = infty will not appear, at
       * leat as far as I'm aware now: f is updated only by
       * open and solved sons; the failed sons are skipped.
       * Hence we have to check here also whether f == -2
       * -- which is the case iff all sons are failed.
       */
      if ( f != -2 ) {
	if ( minf == -2 || LESS( f, minf ) ) {
	  minf = f;
	  bestedge = i_edge;
	  if ( gcmd_line.debug > 2 ) 
	    printf("\nSAM: update minf value to %d", minf);
	}
      } else {
	/* debugging, just to be fool-proof sure
	 */
	if ( num_failed_sons < i_edge->num_out_nodes ) {
	  printf("\nSAM f == -2 but failed sons num != total sons num?\n\n");
	  exit( 1 );
	}
      }

      /* debugging... just to be fool-proof sure...
       */
      if ( num_failed_sons + num_open_sons + num_solved_sons 
	   != i_edge->num_out_nodes ) {
	printf("\nSAM total number of sons counted wrongly?\n\n");
	exit( 1 );
      }

      /* remember whether one of these guys was solved
       * or open.
       *
       * here, the latter is the case already if just one son
       * is open.
       *
       * the former is the case iff:
       *     (all sons are either failed or solved)
       *         (which is equivalent to saying that there
       *          is no "open" son in our sense here)
       *     AND at least one son is solved.
       *
       */
      if ( num_open_sons > 0 ) {
	ex_open = TRUE;
	if ( gcmd_line.debug > 2 ) printf("\nSAM: open = 1");
      }
      if ( num_open_sons == 0 && num_solved_sons > 0 ) {
	ex_solved = TRUE;
	if ( gcmd_line.debug > 2 ) printf("\nSAM: solved = 1");
      }

    } /* loop over all outgoing edges i_edge */



    /* make the "OR" over the state node.
     */
    if ( !ex_open && !ex_solved ) {
      i_bfs->failed = TRUE;
      i_bfs->f = INFINITY;
    } else {
      if ( ex_solved ) {
	i_bfs->solved = TRUE;
      }
      /* debug...
       */
      if ( minf == -2 ) {
	printf("\n\nSAM min not set? ex-open: %d, ex-solved: %d\n\n",
	       ex_open, ex_solved);
	exit( 1 );
      }    
      if ( bestedge == NULL ) {
	printf("\n\nSAM bestedge not set?\n\n");
	exit( 1 );
      }    
      if ( gcmd_line.debug > 2 ) 
	printf("\nSAM: setting new f value %d", 1 + minf);
      i_bfs->f = 1 + minf;
      i_bfs->best_outgoing_edge = bestedge;
    }



    /* go 1 more step towards the root of the tree
     */
    if ( i_bfs->ingoing_edge != NULL ) {
      i_bfs = i_bfs->ingoing_edge->in_node;
    } else {
      break;
    }

  } /* while loop backwards over bfs nodes from open leaf */

}



/* June'11: sufficient test for validity under helpful actions
 * pruning: ask if every unsat leaf in SAM-AO* solution has h=infty
 */
Bool verify_plan( void )

{

  BfsEdge_pointer upperlayer[2048], lowerlayer[2048];
  int num_upperlayer, num_lowerlayer;
  int i, j, t, totalcount = 0, nondetcount = 0;

  if ( gbfs_initial_state->best_outgoing_edge ) {
    lowerlayer[0] = gbfs_initial_state->best_outgoing_edge;
    num_lowerlayer = 1;
  } else {
    num_lowerlayer = 0;
  }

  t = 0;
  while ( num_lowerlayer > 0 ) {
    num_upperlayer = 0;
    for ( i = 0; i < num_lowerlayer; i++ ) {
      for ( j = 0; j < lowerlayer[i]->num_out_nodes; j++ ) {
	if ( j >= 10 ) {
	    printf("\n\ntoo many different outcomes (>10) in plan verify!\n\n");
	    exit( 1 );
	}
	/* June'11: need to ask "!lowerlayer[i]->out_nodes[j]->failed"
	 * here since nodes set as failed by "update" keep a
	 * previously set best_outgoing_edge (and may be leaf nodes in
	 * SAM-AO*)! could have changed above in search but figured
	 * it's better here in case I overlooked any issues that make
	 * it necessary to keep the best_outgoing_edge (though I don't
	 * think so)
	 */
	if ( !lowerlayer[i]->out_nodes[j]->failed &&
	     lowerlayer[i]->out_nodes[j]->best_outgoing_edge != NULL ) {
	  if ( num_upperlayer == 2048 ) {
	    printf("\n\ntoo many actions in one level for plan verify!\n\n");
	    exit( 1 );
	  }
	  upperlayer[num_upperlayer] = 
	    lowerlayer[i]->out_nodes[j]->best_outgoing_edge;
	  num_upperlayer++;
	} else {
	  if ( lowerlayer[i]->out_nodes[j]->failed ) {
	    /* must be a proved-unsat leaf in SAM-AO*
	     *
	     * ask if h = infty here!!!
	     */
	    if ( !(lowerlayer[i]->out_nodes[j]->f == INFINITY) ) {
	      /* sanity debugging test
	       */
	      if ( !gcmd_line.helpful ) {
		printf("\n\nplan verify failed although no helpful actions???\n\n");
		exit(1);
	      }
	      return FALSE;
	    }
	  } else {
	    /* must be a goal state. double-check.
	     */
	    if ( !contains_goal( &(lowerlayer[i]->out_nodes[j]->S) ) ) {
	      printf("\n\nnon-failed leaf state no goal in plan print??\n\n");
	      exit( 1 );
	    }
	  }
	}
      } /* endfor j over out-nodes of edge */
    } /* end loop over edges in lower layer */
    
    /* copy upper into lower
     */
    for ( i = 0; i < num_upperlayer; i++ ) {
      lowerlayer[i] = upperlayer[i];
    }
    num_lowerlayer = num_upperlayer;
    
    t++; 
  } /* end loop over time steps t */

  return TRUE;

}



void print_plan( void )

{

  BfsEdge_pointer upperlayer[2048], lowerlayer[2048];
  int num_upperlayer, lowerlayerson[2048][10], num_lowerlayer;
  int i, j, t, totalcount = 0, nondetcount = 0;

  printf("\n\nff: found plan as follows");
  fflush( stdout );

  if ( gbfs_initial_state->best_outgoing_edge ) {
    lowerlayer[0] = gbfs_initial_state->best_outgoing_edge;
    num_lowerlayer = 1;
  } else {
    num_lowerlayer = 0;
  }

  t = 0;
  while ( num_lowerlayer > 0 ) {
    num_upperlayer = 0;
    for ( i = 0; i < num_lowerlayer; i++ ) {
      for ( j = 0; j < lowerlayer[i]->num_out_nodes; j++ ) {
	if ( j >= 10 ) {
	    printf("\n\ntoo many different outcomes (>10) in plan printing!\n\n");
	    exit( 1 );
	}
	/* June'11: need to ask "!lowerlayer[i]->out_nodes[j]->failed"
	 * here since nodes set as failed by "update" keep a
	 * previously set best_outgoing_edge (and may be leaf nodes in
	 * SAM-AO*)! could have changed above in search but figured
	 * it's better here in case I overlooked any issues that make
	 * it necessary to keep the best_outgoing_edge (though I don't
	 * think so)
	 */
	if ( !lowerlayer[i]->out_nodes[j]->failed &&
	     lowerlayer[i]->out_nodes[j]->best_outgoing_edge != NULL ) {
	  if ( num_upperlayer == 2048 ) {
	    printf("\n\ntoo many actions in one level for plan print!\n\n");
	    exit( 1 );
	  }
	  upperlayer[num_upperlayer] = 
	    lowerlayer[i]->out_nodes[j]->best_outgoing_edge;
	  lowerlayerson[i][j] = num_upperlayer;
	  num_upperlayer++;
	} else {
	  if ( lowerlayer[i]->out_nodes[j]->failed ) {
	    /* must be a proved-unsat leaf in SAM-AO*
	     */
	    lowerlayerson[i][j] = -2;
	  } else {
	    /* must be a goal state. double-check.
	     */
	    if ( !contains_goal( &(lowerlayer[i]->out_nodes[j]->S) ) ) {
	      printf("\n\nnon-failed leaf state no goal in plan print??\n\n");
	      exit( 1 );
	    }
	    lowerlayerson[i][j] = -1;
	  }
	}
      } /* endfor j over out-nodes of edge */
    } /* end loop over edges in lower layer */

    /* print this (lower) layer
     */
    printf("\n------------------------------------------------------------------");
    for ( i = 0; i < num_lowerlayer; i++ ) {
      printf("\n%3d||%d --- ", t, i);
      print_op_name( lowerlayer[i]->op );
      totalcount++;
      if ( lowerlayer[i]->num_out_nodes > 1 ) {
	nondetcount++;
      }
      for ( j = 0; j < lowerlayer[i]->num_out_nodes; j++ ) {
	if ( lowerlayerson[i][j] >= 0 ) {
	  printf(" --- SON %d: %d||%d", 
		 j, t+1, lowerlayerson[i][j]);
	} else {
	  if ( lowerlayerson[i][j] == -1 ) {
	    printf(" --- SON %d: GOAL", j);
	  } else {
	    printf(" --- SON %d: FAIL", j);
	  }	    
	}
      }
    }
    
    /* copy upper into lower
     */
    for ( i = 0; i < num_upperlayer; i++ ) {
      lowerlayer[i] = upperlayer[i];
    }
    num_lowerlayer = num_upperlayer;
    
    t++; 
  } /* end loop over time steps t */

  printf("\n------------------------------------------------------------------");
  printf("\n\ntree layers: %d", t);
  printf("\ntotal nr. actions: %d", totalcount);
  gplanactions = totalcount;
  gplannondetactions = nondetcount;
  gplanlayers = t;

}



void print_used_nondets( BfsNode *n )

{

  int i;

  for ( i = 0; i < n->num_used_nondets; i++ ) {
    print_op_name(n->used_nondets[i]);
    printf("\n");
  }

}














/****************************
 * STATE HANDLING FUNCTIONS *
 ****************************/























Bool contains_goal( State *S )

{

  int i, j;

  for ( i = 0; i < ggoal_state.num_F; i++ ) {
    for ( j = 0; j < S->num_F; j++ ) {
      if ( S->F[j] == ggoal_state.F[i] ) break;
    }
    if ( j == S->num_F ) {
      return FALSE;
    }
  }

  return TRUE;

}



/* function that computes state transition. Adds go before deletes!
 *
 * SAP FF: this computes the outcome for one particulsar EFFECT!
 * ie for one of the nondet-outcomes of thr action.
 *
 * NOTE: the fancy implementation made sense before. not so sure now.
 * anyway, this is not important, leave as is.
 */
void result_to_dest( State *dest, State *source, int ef )

{

  static Bool first_call = TRUE;
  static Bool *in_source, *in_dest, *in_del;
  static int *del, num_del;

  int i, j;

  if ( first_call ) {
    in_source = ( Bool * ) calloc( gnum_ft_conn, sizeof( Bool ) );
    in_dest = ( Bool * ) calloc( gnum_ft_conn, sizeof( Bool ) );
    in_del = ( Bool * ) calloc( gnum_ft_conn, sizeof( Bool ) );
    del = ( int * ) calloc( gnum_ft_conn, sizeof( int ) );
    for ( i = 0; i < gnum_ft_conn; i++ ) {
      in_source[i] = FALSE;
      in_dest[i] = FALSE;
      in_del[i] = FALSE;
    }
    first_call = FALSE;
  }

  /* setup true facts for effect cond evaluation,
   * and evaluate the condition.
   */
  for ( i = 0; i < source->num_F; i++ ) {
    in_source[source->F[i]] = TRUE;
  }
  for ( j = 0; j < gef_conn[ef].num_PC; j++ ) {
    if ( !in_source[gef_conn[ef].PC[j]] ) break;
  }
  /* SAP-FF: if an effect has its condition violated,
   * then the outcome state will simply be unchanged.
   */
  if ( j < gef_conn[ef].num_PC ) {
    source_to_dest( dest, source );
    for ( i = 0; i < source->num_F; i++ ) {
      in_source[source->F[i]] = FALSE;
    }
    return;
  }
  
  /* setup deleted facts
   */
  num_del = 0;
  for ( j = 0; j < gef_conn[ef].num_D; j++ ) {
    if ( in_del[gef_conn[ef].D[j]] ) continue;
    in_del[gef_conn[ef].D[j]] = TRUE;
    del[num_del++] = gef_conn[ef].D[j];
  }

  /* put all non-deleted facts from source into dest.
   * need not check for put-in facts here,
   * as initial state is made doubles-free, and invariant keeps
   * true through the transition procedure
   */
  dest->num_F = 0;
  for ( i = 0; i < source->num_F; i++ ) {
    if ( in_del[source->F[i]] ) {
      continue;
    }
    dest->F[dest->num_F++] = source->F[i];
    in_dest[source->F[i]] = TRUE;
  }

  /* now, finally, add the adds to dest; 
   * each fact at most once!
   */
  for ( j = 0; j < gef_conn[ef].num_A; j++ ) {
    if ( in_dest[gef_conn[ef].A[j]] ) {
      continue;
    }
    dest->F[dest->num_F++] = gef_conn[ef].A[j];
    in_dest[gef_conn[ef].A[j]] = TRUE;
  }

  /* unset infos
   */
  for ( i = 0; i < source->num_F; i++ ) {
    in_source[source->F[i]] = FALSE;
  }
  for ( i = 0; i < dest->num_F; i++ ) {
    in_dest[dest->F[i]] = FALSE;
  }
  for ( i = 0; i < num_del; i++ ) {
    in_del[del[i]] = FALSE;
  }

}



void source_to_dest( State *dest, State *source )

{

  int i;

  for ( i = 0; i < source->num_F; i++ ) {
    dest->F[i] = source->F[i];
  }
  dest->num_F = source->num_F;

}



void copy_source_to_dest( State *dest, State *source )

{

  int i, m;

  if ( dest->max_F < source->num_F ) {
    if ( dest->F ) {
      free( dest->F );
    }
    if ( source->num_F + 50 > gnum_ft_conn ) {
      m = gnum_ft_conn;
    } else {
      m = source->num_F + 50;
    }
    dest->F = ( int * ) calloc( m, sizeof( int ) );
    dest->max_F = m;
  }

  for ( i = 0; i < source->num_F; i++ ) {
    dest->F[i] = source->F[i];
  }
  dest->num_F = source->num_F;

}



void print_state( State S )

{

  int i;
  
  for ( i = 0; i < S.num_F; i++ ) {
    printf("\n");
    print_ft_name( S.F[i] );
  }

}
























/********************************
 * DUPLICATE CHECKING FUNCTIONS *
 ********************************/




















Bool has_duplicate_simple_predecessor( State *S, BfsNode *leaf ) 

{

  BfsNode *i_bfs;
  BfsEdge *i_edge, *bestedge;
  int s;

  s = state_sum( S );

  i_bfs = leaf;
  while ( TRUE ) {
    if ( i_bfs->ingoing_edge == NULL ) {
      break;
    }
    i_bfs = i_bfs->ingoing_edge->in_node;

    if ( state_sum( &(i_bfs->S) ) != s ) {
      continue;
    }

    if ( !same_state( S, &(i_bfs->S) ) ) {
      continue;
    }

    if ( gcmd_line.debug >= 2 ) {
      printf("\nfound duplicate! :");
      print_state( i_bfs->S );
    }
    gnum_simple_duplicates++;
    return TRUE;
  } /* while loop backwards over bfs nodes from open leaf */

  return FALSE;

}



Bool same_state( State *S1, State *S2 ) 

{

  int i, j;

  if ( S2->num_F != S1->num_F ) {
    return FALSE;
  }

  for ( i = 0; i < S2->num_F; i++ ) {
    for ( j = 0; j < S1->num_F; j++ ) {
      if ( S1->F[j] == S2->F[i] ) {
	break;
      }
    }
    if ( j == S1->num_F ) {
      return FALSE;
    }
  }

  return TRUE;

}



int state_sum( State *S )

{

  int i, sum = 0;

  for ( i = 0; i < S->num_F; i++ ) {
    sum += gft_conn[S->F[i]].rand;
  }

  return sum;

}


